/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
* (OS/2 compatibility (c) Jay Schwartz, Elipse Ltd.)                    *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* Historical note: in the original DOS version of NLED, the Datalight   *
* compiler (subsequently becoming the Zortech compiler) was used. Its   *
* "qkdisp" library was initially used for screen I/O, but problems with *
* it caused me to learn assembly language so that I could rewrite it to *
* my performance requirements. (Hey, I was working on an original 4.77  *
* MHz IBM PC with a CGA adapter - performance mattered). Later ports of *
* NLED to non-DOS platforms were done mostly by just recoding those     *
* routines (in C) with platform specific code. Since the DOS assembly   *
* language code has Datalight (Northwest Software) code in it, and      *
* since the DOS version is essentially dead, the modified QKDISP.ASM    *
* has not been released as part of the GPLing of NLED. Just in case you *
* were wondering why the naming conventions do not match the rest of    *
* NLED, and why there are some empty functions!                         *
*                                                                       *
* File: QKDISP.C                         Date Last Modified: 05-May-03  *
*                                                                       *
* Quick display functions for UNIX and Windows 95 (based very loosely   *
* on the library DISP.ASM by Northwest Software). Allows simulation of  *
* direct control of PC-compatible video hardware from Zortech C.        *
*                                                                       *
* 08-Jul-96: Win95 port                                                 *
* 09-Jan-97: OS/2  port                                                 *
* 02-Feb-02: Borland port for cleaner Windows code                      *
* 06-Feb-02: Fixed AIX problems, improved Win95(MS) code                *
* 09-Feb-02: Improved Win95 code some more                              *
* 21-Mar-03: removed Win95 requirement to run without scrollbars        *
* 05-May-03: Borland kludge in qd_open to handle scrollbars             *
*                                                                       *
* Routines are:                                                         *
*   qd_beep() - make a beep sound                                       *
*   qd_boldon() - turn bold (highlighted) mode on                       *
*   qd_boldoff() - turn bold (highlighted) mode off                     *
*   qd_clear() - clear screen                                           *
*   qd_close() - deinitialize display routines                          *
*   qd_color() - set screen colors                                      *
*   qd_cols() - get number of actual screen columns                     *
*   qd_eeol() - erase from current cursor position to end of line       *
*   qd_flush() - bring cursor up to date                                *
*   qd_getch() - get keystroke                                          *
*   qd_move() - move cursor to a new position                           *
*   qd_off() - turn display off                                         *
*   qd_on() - turn display back on                                      *
*   qd_open() - initialize display routines                             *
*   qd_puts() - write as string to current cursor position              *
*   qd_revon() - turn reverse video mode on                             *
*   qd_revoff() - turn reverse video mode off                           *
*   qd_rows() - get number of actual screen rows                        *
*   qd_scrollup() - scroll part of screen up one line                   *
*   qd_scrolldown() - scroll part of screen down one line               *
*   qd_snowoff() - tell other routines display is not subject to snow   *
*   qd_snowon() - tell other routines display is subject to snow        *
*   qd_usebios() - force routines to use ROM BIOS to write to screen    *
*   qd_saveterm() - save current tty settings                           *
*   qd_resetterm() - set tty to normal terminal state                   *
*   qd_fixterm() - restore previously saved tty settings                *
\***********************************************************************/

/************************************************************************\
* IMPORTANT - you will have to MANUALLY set ONLY ONE of these to 1.      *
*  These settings are independent, but related to, similar settings in   *
*  NLED.H. In particular, both WIN95 and BORLAND here match WIN95 there. *
*  WIN95 really means MS VC++, while BORLAND is a less desirable Windows *
*  alternative (less desirable because the port isn't very good).        *
\************************************************************************/
#define WIN95   0
#define BORLAND 0
#define __OS2__ 0
#define LINUX   1
#define AIX     0

#if BORLAND
#include <conio.h>
    short qd_normattr = 0x07, /* normal, ...      */
	  qd_boldattr = 0x0f, /* ...bold, and...  */
	  qd_revattr  = 0x70; /* ...reverse video */
    short qd_numcols = 0,
	  qd_numrows = 0;
#elif WIN95
#include <windows.h>
#include "nledkeys.h"
    COORD qd_coord;           /* current screen position */
    short qd_normattr = 0x07, /* normal, ...      */
	  qd_boldattr = 0x0f, /* ...bold, and...  */
	  qd_revattr  = 0x70, /* ...reverse video */
	  qd_saveattr = 0x07, /* save original attribute */
	  qd_clrattr  = 0x07; /* used to clear to end of line */
    short qd_numcols = 0,
	  qd_numrows = 0,
	  qd_drawtop = -1,    /* line to draw (-1: none, qd_numrows: all) */
	  qd_drawcol = -1,    /* first column to redraw if one line only */
	  qd_drawbot = 0;     /* draw status (bottom) line? */
    HANDLE qd_out, qd_back,   /* output and backbuffer consoles */
	   qd_orig;           /* original console (to restore at end) */
    CHAR_INFO *qd_buff = NULL;
#elif __OS2__
#define INCL_VIO
#include <os2.h>
#include <string.h>
#include <stdio.h>
#include <conio.h>
    typedef struct
	{
	short X;
	short Y;
	} COORD;
    HVIO hvio = (HVIO) 0;
    COORD qd_coord;           /* current screen position */
    short qd_normattr = 0x07, /* normal, ...      */
	  qd_boldattr = 0x0f, /* ...bold, and...  */
	  qd_revattr  = 0x70, /* ...reverse video */
	  qd_saveattr = 0x07, /* save original attribute */
	  qd_currattr = 0x07; /* current screen attribute */
    short qd_numcols = 0,
	  qd_numrows = 0;
#elif LINUX
#include <ncurses.h>
#else
#include <curses.h>
#endif

#ifdef beep
#undef beep
#endif

#if WIN95
/* assuming we are drawing on line "row", set redrawing flags accordingly */
void touchline(short row, short col)
{
    if (row == -1) {                /* request for whole screen */
	qd_drawtop = qd_numrows;    /* ..the whole top section */
	qd_drawcol = -1;
	qd_drawbot = 1;             /*.. and the status line */
    }
    else if (row == qd_numrows - 1) /* is the status line */
	qd_drawbot = 1;
    else if (qd_drawtop == -1) {    /* is the first time drawing in the top */
	qd_drawtop = row;           /* ..just redraw this line */
	qd_drawcol = col;
    }
    else if (qd_drawtop != row) {   /* at least one other line in the top */
	qd_drawtop = qd_numrows;    /* ..so do the whole top */
	qd_drawcol = -1;
    }
    else if (col < qd_drawcol)      /* same line as before but left of prev */
	qd_drawcol = -1;
}
#endif

void qd_beep(void)
{
#if BORLAND
    putch(7);
#elif WIN95
    MessageBeep(MB_OK);
#elif __OS2__
    _putch( 7 );
#else
    putchar(7);
#endif
}

void qd_boldon(void)
{
#if BORLAND
    textattr(qd_boldattr);
#elif WIN95
    SetConsoleTextAttribute(qd_back, qd_boldattr);
    qd_clrattr = qd_normattr;
#elif __OS2__
    qd_currattr = qd_boldattr;
#else
    attron(A_BOLD);
#endif
}

void qd_boldoff(void)
{
#if BORLAND
    textattr(qd_normattr);
#elif WIN95
    SetConsoleTextAttribute(qd_back, qd_normattr);
    qd_clrattr = qd_normattr;
#elif __OS2__
    qd_currattr = qd_normattr;
#else
    attroff(A_BOLD);
#endif
}

void qd_clear(void)
{
#if BORLAND
    clrscr();
#elif WIN95
    DWORD junk;
    qd_coord.X = qd_coord.Y = 0;
    FillConsoleOutputCharacter(qd_back, ' ', qd_numrows * qd_numcols,
     qd_coord, &junk);
    FillConsoleOutputAttribute(qd_back, qd_normattr, qd_numrows * qd_numcols,
     qd_coord, &junk);
    SetConsoleCursorPosition(qd_back, qd_coord);
    touchline(-1, -1);
#elif __OS2__
    char str[4] = { ' ', 0, 0, 0};
    str[1] = qd_normattr;
    qd_coord.X = qd_coord.Y = 0;
    VioWrtNCell( str, qd_numrows * qd_numcols, qd_coord.Y, qd_coord.X, hvio );
#else
    clear();
#endif
}

void qd_close(void)
{
#if BORLAND
     gotoxy(1, qd_numrows);
     normvideo();
     _setcursortype(_NORMALCURSOR);
     clreol();
#elif WIN95
    DWORD junk;
    extern void qd_flush(void);
    touchline(-1, -1);
    qd_flush();

    /*  old Windows code that moved down one line before exiting:

	qd_coord.X = 0;
	qd_coord.Y = qd_numrows - 1;
	FillConsoleOutputAttribute(qd_out, qd_saveattr, qd_numcols, qd_coord,
	 &junk);
	SetConsoleTextAttribute(qd_out, qd_saveattr);
	SetConsoleCursorPosition(qd_out, qd_coord);
    */

    SetConsoleActiveScreenBuffer(qd_orig);
    CloseHandle(qd_out);
    CloseHandle(qd_back);
    free(qd_buff);
#elif __OS2__
#else
    endwin();
#endif
}

void qd_color(int color)
{
#if BORLAND || WIN95 || __OS2__
    qd_normattr = color;
    qd_boldattr = color | 0x08;
    qd_revattr  = ((color & 0x07) << 4) | ((color & 0x070) >> 4);
#endif
#if BORLAND
    textattr(qd_normattr);
#elif WIN95
    qd_clrattr = qd_normattr;
    SetConsoleTextAttribute(qd_out, qd_normattr);
    SetConsoleTextAttribute(qd_back, qd_normattr);
#endif
}

int qd_cols(void)
{
#if WIN95 || __OS2__ || BORLAND
    return qd_numcols;
#else
    return COLS;
#endif
}

void qd_eeol(void)
{
#if BORLAND
    clreol();
#elif WIN95
    DWORD junk;
    FillConsoleOutputCharacter(qd_back, ' ', qd_numcols - qd_coord.X,
     qd_coord, &junk);
    FillConsoleOutputAttribute(qd_back, qd_clrattr, qd_numcols - qd_coord.X, 
     qd_coord, &junk);
    touchline(qd_coord.Y, qd_coord.X);
#elif __OS2__
    char str[4] = { ' ', 0, 0, 0};
    str[1] = qd_currattr;
    VioWrtNCell( str, qd_numcols - qd_coord.X, qd_coord.Y, qd_coord.X, hvio );
#else
    clrtoeol();
#endif
}

void qd_fixterm(void)
{
#if WIN95
    COORD c, orig;
    SMALL_RECT r;
    CONSOLE_SCREEN_BUFFER_INFO scr;

    /* copy screen into backbuffer */
    c.X = qd_numcols;
    c.Y = qd_numrows;
    orig.X = r.Left = 0;
    orig.Y = r.Top = 0;
    r.Right = qd_numcols - 1;
    r.Bottom = qd_numrows - 1;
    ReadConsoleOutput(qd_out, qd_buff, c, orig, &r);
    WriteConsoleOutput(qd_back, qd_buff, c, orig, &r);

    GetConsoleScreenBufferInfo(qd_out, &scr);
    qd_coord = scr.dwCursorPosition;
    SetConsoleCursorPosition(qd_back, qd_coord);
#elif __OS2__ || BORLAND
#else
    fixterm();
#endif
}

void qd_flush(void)
{
#if WIN95
    COORD c, orig;
    SMALL_RECT r;

    /* cut data from backbuffer and place it on screen console */
    if (qd_drawtop == qd_numrows && qd_drawbot) {
	c.X = qd_numcols;
	c.Y = qd_numrows;
	orig.X = orig.Y = 0;
	r.Left = 0;
	r.Right = qd_numcols - 1;
	r.Top = 0;
	r.Bottom = qd_numrows - 1;
	ReadConsoleOutput(qd_back, qd_buff, c, orig, &r);
	WriteConsoleOutput(qd_out, qd_buff, c, orig, &r);
    }
    else {
	if (qd_drawbot) {
	    c.X = qd_numcols;
	    c.Y = qd_numrows;
	    orig.X = r.Left = 0;
	    orig.Y = r.Top = qd_numrows - 1;
	    r.Right = qd_numcols - 1;
	    r.Bottom = qd_numrows - 1;
	    ReadConsoleOutput(qd_back, qd_buff, c, orig, &r);
	    WriteConsoleOutput(qd_out, qd_buff, c, orig, &r);
	}
	if (qd_drawtop == qd_numrows) {
	    c.X = qd_numcols;
	    c.Y = qd_numrows;
	    orig.X = r.Left = 0;
	    orig.Y = r.Top = 0;
	    r.Right = qd_numcols - 1;
	    r.Bottom = qd_numrows - 2;
	    ReadConsoleOutput(qd_back, qd_buff, c, orig, &r);
	    WriteConsoleOutput(qd_out, qd_buff, c, orig, &r);
	}
	else if (qd_drawtop != -1) {
	    c.X = qd_numcols;
	    c.Y = qd_numrows;
	    orig.X = r.Left = (qd_drawcol == -1 ? 0 : qd_drawcol);
	    orig.Y = r.Top = qd_drawtop;
	    r.Right = qd_numcols - 1;
	    r.Bottom = qd_drawtop;
	    ReadConsoleOutput(qd_back, qd_buff, c, orig, &r);
	    WriteConsoleOutput(qd_out, qd_buff, c, orig, &r);
	}
    }
    qd_drawtop = qd_drawcol = -1;
    qd_drawbot = 0;

    SetConsoleCursorPosition(qd_out, qd_coord);
#elif __OS2__ || BORLAND
#else
    refresh();
#endif
}

int qd_getch(void)
{
#if BORLAND
    int c;
    _setcursortype(_NORMALCURSOR);
    c = getch();
    c = c > 0 ? c : getch() | 0xFF00;
    _setcursortype(_NOCURSOR);
    return c;
#elif WIN95
    int c;
    c = _getch();
    if (c == 224 || c == 0) {
	c = _getch() | 0xFF00;
	if (c == 0xFF86)
		c = 0xFF84; /* ctrl PgUp has changed */
    }
    return c;
#elif __OS2__
    int ch = _getch();
    if ((ch > 0) && (ch < 128))
	return ch;
    else
	return  _getch( ) | 0xff00;
#else
    return getch();
#endif
}

void qd_move(int row, int col)
{
#if BORLAND
    gotoxy(col + 1, row + 1);
#elif WIN95
    qd_coord.X = col;
    qd_coord.Y = row;
    SetConsoleCursorPosition(qd_back, qd_coord);
#elif __OS2__
    VioSetCurPos( row, col, hvio );
    qd_coord.X = col;
    qd_coord.Y = row;
#else
    move(row, col);
#endif
}

void qd_off(void)
{
}

void qd_on(void)
{
}

void qd_open(void)
{
#if BORLAND
    struct text_info ti;
    system("cls");  /* kludge so when running in a window with scrollbars
		       the screen doesn't end up a mess at the end */
    gettextinfo(&ti);
    qd_numcols = ti.screenwidth;
    qd_numrows = ti.screenheight;
#elif WIN95
    DWORD mode;
    CONSOLE_SCREEN_BUFFER_INFO scr;
    CONSOLE_CURSOR_INFO cursinfo;
    COORD c, orig;
    SMALL_RECT r;

    qd_orig = GetStdHandle(STD_OUTPUT_HANDLE);
    qd_out = CreateConsoleScreenBuffer(GENERIC_WRITE | GENERIC_READ,
     0, NULL, CONSOLE_TEXTMODE_BUFFER, NULL);
    qd_back = CreateConsoleScreenBuffer(GENERIC_WRITE | GENERIC_READ,
     0, NULL, CONSOLE_TEXTMODE_BUFFER, NULL);
    if (qd_orig && qd_out && qd_back) {
	GetConsoleCursorInfo(qd_orig, &cursinfo);
	GetConsoleScreenBufferInfo(qd_orig, &scr);
	c.Y = qd_numrows = scr.srWindow.Bottom - scr.srWindow.Top + 1;
	c.X = qd_numcols = scr.srWindow.Right - scr.srWindow.Left + 1;

	SetConsoleScreenBufferSize(qd_out, c);
	SetConsoleScreenBufferSize(qd_back, c);

	SetConsoleActiveScreenBuffer(qd_out);

	SetConsoleTitle("NLED");
	SetConsoleMode(qd_out, 0);
	SetConsoleMode(qd_back, 0);
	cursinfo.bVisible = FALSE;
	SetConsoleCursorInfo(qd_back, &cursinfo);

	qd_saveattr = scr.wAttributes;
	qd_buff = malloc(sizeof(CHAR_INFO) * qd_numrows * qd_numcols);
	qd_fixterm();

	/* In fullscreen mode only, fixterm turns off qd_out's */
	/* cursor, so this MUST be done down here !?!?!        */
	cursinfo.bVisible = TRUE;
	SetConsoleCursorInfo(qd_out, &cursinfo);
    }
#elif __OS2__
    VIOMODEINFO mdata;
    memset( &mdata, '\0', sizeof( mdata ) );
    mdata.cb = sizeof( mdata );
    hvio = 0;
    VioGetMode( &mdata, hvio );
    qd_numcols = mdata.col;
    qd_numrows = mdata.row;
#else
    initscr();
    noecho();
    raw();  /* instead of cbreak() so that Ctrl/C won't break */
    keypad(stdscr,1);
    scrollok(stdscr,1);
#endif
}

void qd_puts(int length, char *string)
{
#if BORLAND
    int i;
    for (i = 0; i < length; i++)
	switch(string[i]) {
	  case 7:
	  case 8:
	  case 13:
		putch('.');
		break;
	  default:
		putch(string[i]);
	}
#elif WIN95
    DWORD junk;
    int i;
    char c;
    for (i = 0; i < length; i++) {
	switch(c = string[i]) {
	  case 7:
	  case 8:
	  case 13:
		c = '.';
	}
	WriteConsole(qd_back, &c, 1, &junk, NULL);
    }
    touchline(qd_coord.Y, qd_coord.X);
    qd_coord.X += length;
#elif __OS2__
    char str[4] = { 0, 0, 0, 0};
    str[0] = qd_currattr;
    VioWrtCharStr( string, length, qd_coord.Y, qd_coord.X, hvio );
    VioWrtNAttr( str, length, qd_coord.Y, qd_coord.X, hvio );
    qd_coord.X += length;
#else
    int i;
    for (i = 0; i < length; i++)
	addch((string[i] >= ' ' && string[i] <= '~') ? string[i] : '.');
#endif
}

void qd_resetterm(void)
{
#if BORLAND
    _setcursortype(_NORMALCURSOR);
#elif WIN95
    qd_flush();
    SetConsoleTextAttribute(qd_out, qd_normattr);
    SetConsoleCursorPosition(qd_out, qd_coord);
#elif __OS2__
#else
    resetterm();
#endif
}

void qd_revon(void)
{
#if BORLAND
    textattr(qd_revattr);
#elif WIN95
    SetConsoleTextAttribute(qd_back, qd_revattr);
    qd_clrattr = qd_revattr;
#elif __OS2__
    qd_currattr = qd_revattr;
#else
    attron(A_REVERSE);
#endif
}

void qd_revoff(void)
{
#if BORLAND
    textattr(qd_normattr);
#elif WIN95
    SetConsoleTextAttribute(qd_back, qd_normattr);
    qd_clrattr = qd_normattr;
#elif __OS2__
    qd_currattr = qd_normattr;
#else
    attroff(A_REVERSE);
#endif
}

int qd_rows(void)
{
#if WIN95 || __OS2__ || BORLAND
    return qd_numrows;
#else
    return LINES;
#endif
}

void qd_saveterm(void)
{
#if WIN95 || __OS2__ || BORLAND
#else
    saveterm();
#endif
}

void qd_scrollup(int startrow, int endrow)
{
#if BORLAND
    movetext(1, startrow + 2, qd_numcols, endrow + 1, 1, startrow + 1);
#elif WIN95
    SMALL_RECT tot, scr;
    COORD dest;
    CHAR_INFO ch;
    tot.Left = 0;
    tot.Right = qd_numcols - 1;
    tot.Top = startrow;
    tot.Bottom = endrow;
    scr.Left = 0;
    scr.Right = qd_numcols - 1;
    scr.Top = startrow + 1;
    scr.Bottom = endrow;
    dest.X = 0;
    dest.Y = startrow;
    ch.Char.AsciiChar = ' ';
    ch.Attributes = qd_normattr;
    ScrollConsoleScreenBuffer(qd_back, &scr, &tot, dest, &ch);
    touchline(startrow, 0);
    touchline(endrow, 0);
#elif __OS2__
    char str[4] = { ' ', 0, 0, 0};
    str[1] = qd_normattr;
    VioScrollUp( startrow, 0, endrow, qd_numcols - 1 , 1, str, hvio );
#else
    setscrreg(startrow, endrow);
    scrl(1);
    setscrreg(0, LINES - 1);

    /* old AIX code:
      WINDOW *sub, *sub1;
      sub1 = subwin(stdscr, endrow-startrow+1, COLS, startrow, 0);
      sub = newwin(endrow - startrow + 1, COLS, startrow, 0);
      idlok(sub, 1);
      overwrite(sub1, sub);
      delwin(sub1);
      wmove(sub, 0, 0);
      wdeleteln(sub);
      overwrite(sub, stdscr);
      delwin(sub);
    */
#endif
}

void qd_scrolldown(int startrow, int endrow)
{
#if BORLAND
    movetext(1, startrow + 1, qd_numcols, endrow, 1, startrow + 2);
#elif WIN95
    SMALL_RECT tot, scr;
    COORD dest;
    CHAR_INFO ch;
    tot.Left = 0;
    tot.Right = qd_numcols - 1;
    tot.Top = startrow;
    tot.Bottom = endrow;
    scr.Left = 0;
    scr.Right = qd_numcols - 1;
    scr.Top = startrow;
    scr.Bottom = endrow - 1;
    dest.X = 0;
    dest.Y = startrow + 1;
    ch.Char.AsciiChar = ' ';
    ch.Attributes = qd_normattr;
    ScrollConsoleScreenBuffer(qd_back, &scr, &tot, dest, &ch);
    touchline(startrow, 0);
    touchline(endrow, 0);
#elif __OS2__
    char str[4] = { ' ', 0, 0, 0};
    str[1] = qd_normattr;
    VioScrollDn( startrow, 0, endrow, qd_numcols - 1, 1, str, hvio );
#else
    setscrreg(startrow, endrow);
    scrl(-1);
    setscrreg(0, LINES - 1);

    /* old AIX code:
      WINDOW *sub, *sub1;
      sub1 = subwin(stdscr, endrow - startrow + 1, COLS, startrow, 0);
      sub = newwin(endrow - startrow + 1, COLS, startrow, 0);
      idlok(sub, 1);
      overwrite(sub1, sub);
      delwin(sub1);
      wmove(sub, 0, 0);
      winsertln(sub);
      overwrite(sub, stdscr);
      delwin(sub);
    */
#endif
}

void qd_snowoff(void)
{
}

void qd_snowon(void)
{
}

void qd_usebios(void)
{
}
